# noqa: INP001
"""Unit and fuzz tests for ``ozi-fix`` utility script"""
# Part of ozi.
# See LICENSE.txt in the project root for details.
import argparse
import os
import pathlib
import typing
from copy import deepcopy
from datetime import timedelta

import pytest
from hypothesis import given, settings
from hypothesis import strategies as st

import ozi.fix
import ozi.new
from ozi.assets import required_pkg_info_patterns
from ozi.assets.structure import root_files, source_files, test_files
from ozi.fix import env

bad_namespace = argparse.Namespace(
    strict=False,
    verify_email=True,
    name='OZI-phony',
    keywords='foo,bar,baz',
    maintainer=[],
    maintainer_email=[],
    author=['foo'],
    author_email=['noreply@oziproject.dev'],
    home_page='foobar',
    summary='A' * 512,
    copyright_head='',
    license_expression='CC0-1.0',
    license_file='LICENSE.txt',
    license='CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
    license_id='ITWASAFICTION',
    license_exception_id='WEMADEITUP',
    topic=['Utilities'],
    status=['7 - Inactive'],
    environment=['Other Environment'],
    framework=['Pytest'],
    audience=['Other Audience'],
    ci_provider='github',
    project_url=['A' * 33 + ', https://oziproject.dev'],
    fix='',
    add=['ozi.phony'],
    remove=['ozi.phony'],
    dist_requires=[],
    allow_file=[],
)


@pytest.fixture()
def bad_project(tmp_path_factory: pytest.TempPathFactory) -> pathlib.Path:
    """Fixture to wrap the ``ozi-new project`` functionality."""
    fn = tmp_path_factory.mktemp('project_')
    namespace = deepcopy(bad_namespace)
    namespace.target = fn
    with pytest.warns(RuntimeWarning):
        ozi.new.project(namespace)
    return fn


@pytest.mark.parametrize('key', required_pkg_info_patterns)
def test_report_missing_required(bad_project: pathlib.Path, key: str) -> None:
    """Check that we warn on missing requirements"""
    with bad_project.joinpath('PKG-INFO').open() as f:
        content = f.read()
    with bad_project.joinpath('PKG-INFO').open('w') as f:
        content = content.replace(key, '')
        f.write(content)
    with pytest.raises(RuntimeWarning):
        ozi.fix.report_missing(bad_project)


@pytest.mark.parametrize('key', [i for i in root_files if i not in ['PKG-INFO']])
def test_report_missing_required_root_file(bad_project: pathlib.Path, key: str) -> None:
    """Check that we warn on missing files"""
    os.remove(bad_project.joinpath(key))
    with pytest.warns(RuntimeWarning):
        ozi.fix.report_missing(bad_project)


@pytest.mark.parametrize('key', ['PKG-INFO'])
def test_report_missing_required_pkg_info_file(bad_project: pathlib.Path, key: str) -> None:
    """Check that we warn on missing files"""
    os.remove(bad_project.joinpath(key))
    with pytest.raises((SystemExit, RuntimeWarning)):
        ozi.fix.report_missing(bad_project)


@pytest.mark.parametrize('key', test_files)
def test_report_missing_required_test_file(bad_project: pathlib.Path, key: str) -> None:
    """Check that we warn on missing files"""
    os.remove(bad_project.joinpath('tests') / key)
    with pytest.warns(RuntimeWarning):
        ozi.fix.report_missing(bad_project)


@pytest.mark.parametrize('key', source_files)
def test_report_missing_required_source_file(bad_project: pathlib.Path, key: str) -> None:
    """Check that we warn on missing files"""
    os.remove(bad_project.joinpath('ozi_phony') / key)
    with pytest.warns(RuntimeWarning):
        ozi.fix.report_missing(bad_project)


@given(
    type=st.just('target'),
    target=st.text(),
    operation=st.text(),
    sources=st.lists(st.text()),
    subdir=st.just(''),
    target_type=st.just('executable'),
)
def test_fuzz_RewriteCommand(  # noqa: DC102
    type: str,
    target: str,
    operation: str,
    sources: typing.List[str],
    subdir: str,
    target_type: str,
) -> None:
    ozi.fix.RewriteCommand(
        type=type,
        target=target,
        operation=operation,
        sources=sources,
        subdir=subdir,
        target_type=target_type,
    )


@given(
    target=st.just('.'),
    name=st.text(),
    fix=st.sampled_from(('test', 'source', 'root')),
    commands=st.lists(st.dictionaries(keys=st.text(), values=st.text())),
)
def test_fuzz_Rewriter(  # noqa: DC102
    target: str, name: str, fix: str, commands: typing.List[typing.Dict[str, str]]
) -> None:
    ozi.fix.Rewriter(target=target, name=name, fix=fix, commands=commands)


@pytest.mark.parametrize('fix', ['test', 'root', 'source'])
def test_Rewriter_bad_project__iadd__dir_nested_warns(  # noqa: DC102
    bad_project: pytest.FixtureRequest, fix: str
) -> None:
    env.globals = env.globals | {'project': vars(bad_namespace)}
    rewriter = ozi.fix.Rewriter(target=str(bad_project), name='ozi_phony', fix=fix)
    with pytest.warns(RuntimeWarning):
        rewriter += ['foo/foo/baz/']


@pytest.mark.parametrize('fix', ['test', 'root', 'source'])
def test_Rewriter_bad_project__iadd__dir(  # noqa: DC102
    bad_project: pytest.FixtureRequest, fix: str
) -> None:
    env.globals = env.globals | {'project': vars(bad_namespace)}
    rewriter = ozi.fix.Rewriter(target=str(bad_project), name='ozi_phony', fix=fix)
    rewriter += ['foo/']
    assert len(rewriter.commands) == 1


def test_Rewriter_bad_project__iadd__bad_fix(  # noqa: DC102
    bad_project: pytest.FixtureRequest,
) -> None:
    env.globals = env.globals | {'project': vars(bad_namespace)}
    rewriter = ozi.fix.Rewriter(target=str(bad_project), name='ozi_phony', fix='')
    with pytest.warns(RuntimeWarning):
        rewriter += ['foo/']
    assert len(rewriter.commands) == 0


def test_Rewriter_bad_project__isub__bad_fix(  # noqa: DC102
    bad_project: pytest.FixtureRequest,
) -> None:
    env.globals = env.globals | {'project': vars(bad_namespace)}
    rewriter = ozi.fix.Rewriter(target=str(bad_project), name='ozi_phony', fix='')
    rewriter -= ['foo.py']
    assert len(rewriter.commands) == 1


@pytest.mark.parametrize('fix', ['test', 'root', 'source'])
def test_Rewriter_bad_project__isub__non_existing_child(  # noqa: DC102
    bad_project: pytest.FixtureRequest, fix: str
) -> None:
    env.globals = env.globals | {'project': vars(bad_namespace)}
    rewriter = ozi.fix.Rewriter(target=str(bad_project), name='ozi_phony', fix=fix)
    with pytest.raises(RuntimeWarning):
        rewriter -= ['foo/']
    assert len(rewriter.commands) == 0


@pytest.mark.parametrize('fix', ['test', 'root', 'source'])
def test_Rewriter_bad_project__isub__child(  # noqa: DC102
    bad_project: pytest.FixtureRequest, fix: str
) -> None:
    env.globals = env.globals | {'project': vars(bad_namespace)}
    rewriter = ozi.fix.Rewriter(target=str(bad_project), name='ozi_phony', fix=fix)
    if fix == 'root':
        pathlib.Path(str(bad_project), 'foo').mkdir()
    elif fix == 'source':
        pathlib.Path(str(bad_project), 'ozi_phony', 'foo').mkdir()
    elif fix == 'test':
        pathlib.Path(str(bad_project), 'tests', 'foo').mkdir()
    rewriter -= ['foo/']
    assert len(rewriter.commands) == 1


@pytest.mark.parametrize('fix', ['test', 'root', 'source'])
def test_Rewriter_bad_project__isub__python_file(  # noqa: DC102
    bad_project: pytest.FixtureRequest, fix: str
) -> None:
    env.globals = env.globals | {'project': vars(bad_namespace)}
    rewriter = ozi.fix.Rewriter(target=str(bad_project), name='ozi_phony', fix=fix)
    if fix == 'root':
        pathlib.Path(str(bad_project), 'foo.py').touch()
    elif fix == 'source':
        pathlib.Path(str(bad_project), 'ozi_phony', 'foo.py').touch()
    elif fix == 'test':
        pathlib.Path(str(bad_project), 'tests', 'foo.py').touch()
    rewriter -= ['foo.py']
    assert len(rewriter.commands) == 1


@pytest.mark.parametrize('fix', ['test', 'root', 'source'])
def test_Rewriter_bad_project__isub__file(  # noqa: DC102
    bad_project: pytest.FixtureRequest, fix: str
) -> None:
    env.globals = env.globals | {'project': vars(bad_namespace)}
    rewriter = ozi.fix.Rewriter(target=str(bad_project), name='ozi_phony', fix=fix)
    if fix == 'root':
        pathlib.Path(str(bad_project), 'foo').touch()
    elif fix == 'source':
        pathlib.Path(str(bad_project), 'ozi_phony', 'foo').touch()
    elif fix == 'test':
        pathlib.Path(str(bad_project), 'tests', 'foo').touch()
    rewriter -= ['foo']
    assert len(rewriter.commands) == 1


@pytest.mark.parametrize('fix', ['test', 'root', 'source'])
def test_Rewriter_bad_project__iadd__file(  # noqa: DC102
    bad_project: pytest.FixtureRequest, fix: str
) -> None:
    env.globals.update({'project': vars(bad_namespace)})
    rewriter = ozi.fix.Rewriter(target=str(bad_project), name='ozi_phony', fix=fix)
    rewriter += ['foo.py']
    assert len(rewriter.commands) == 1


@pytest.mark.parametrize('fix', ['test', 'root', 'source'])
def test_Rewriter_bad_project__iadd__file_from_template(  # noqa: DC102
    bad_project: pytest.FixtureRequest, fix: str
) -> None:
    env.globals.update({'project': vars(bad_namespace)})
    rewriter = ozi.fix.Rewriter(target=str(bad_project), name='ozi_phony', fix=fix)
    pathlib.Path(bad_project / 'templates').mkdir()
    pathlib.Path(bad_project / 'templates' / 'foo.py').touch()
    pathlib.Path(bad_project / 'templates' / 'source').mkdir()
    pathlib.Path(bad_project / 'templates' / 'source' / 'foo.py').touch()
    pathlib.Path(bad_project / 'templates' / 'test').mkdir()
    pathlib.Path(bad_project / 'templates' / 'test' / 'foo.py').touch()
    rewriter += ['foo.py']
    assert len(rewriter.commands) == 1


@pytest.mark.parametrize('fix', ['test', 'root', 'source'])
def test_Rewriter_bad_project__iadd__non_python_file(  # noqa: DC102
    bad_project: pytest.FixtureRequest, fix: str
) -> None:
    env.globals.update({'project': vars(bad_namespace)})
    rewriter = ozi.fix.Rewriter(target=str(bad_project), name='ozi_phony', fix=fix)
    rewriter += ['foo']
    assert len(rewriter.commands) == 1


header = """.. OZI
  Classifier: License-Expression :: Apache-2.0 WITH LLVM-exception
  Classifier: License-File :: LICENSE.txt
"""


def test_preprocess_warns_non_existing_target() -> None:  # noqa: DC102
    namespace = deepcopy(bad_namespace)
    namespace.target = 'temp/foobar'
    with pytest.warns(RuntimeWarning):
        ozi.fix.preprocess(namespace)


def test_preprocess_warns_file_target() -> None:  # noqa: DC102
    namespace = deepcopy(bad_namespace)
    namespace.target = 'temp/foobar.txt'
    pathlib.Path(namespace.target).touch()
    with pytest.warns(RuntimeWarning):
        ozi.fix.preprocess(namespace)


def test_preprocess_existing_target() -> None:  # noqa: DC102
    namespace = deepcopy(bad_namespace)
    namespace.target = '..'
    namespace = ozi.fix.preprocess(namespace)
    assert 'ozi.phony' not in namespace.add
    assert 'ozi.phony' not in namespace.remove


@given(add_items=st.lists(st.text()), remove_items=st.lists(st.text()))
def test_fuzz_preprocess_existing_target(  # noqa: DC102
    add_items: typing.List[str], remove_items: typing.List[str]
) -> None:
    namespace = deepcopy(bad_namespace)
    namespace.add.extend(add_items)
    namespace.remove.extend(remove_items)
    namespace.target = '.'
    namespace = ozi.fix.preprocess(namespace)
    assert 'ozi.phony' not in namespace.add
    assert 'ozi.phony' not in namespace.remove


@settings(deadline=timedelta(milliseconds=500))
@given(payload=st.text(max_size=65535).map(header.__add__), as_message=st.booleans())
def test_fuzz_pkg_info_extra(payload: str, as_message: bool) -> None:  # noqa: DC102
    ozi.fix.pkg_info_extra(payload=payload, as_message=as_message)


@given(s=st.from_regex(r'^([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9._-]*[A-Za-z0-9])$'))
def test_fuzz_underscorify(s: str) -> None:  # noqa: DC102
    ozi.fix.underscorify(s=s)



"""Core OZI packaging management plane module."""
# ozi/__init__.py
# Part of the OZI Project, under the Apache License v2.0 with LLVM Exceptions.
# See LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
from importlib.metadata import version
from typing import List
from warnings import warn

from packaging.version import Version, parse

from .assets import (
    implementation_support,
    meson_min_version,
    metadata_version,
    python_support,
    specification_version,
)

metadata = {
    'ozi': {
        'version': version('OZI'),
        'metadata_version': metadata_version,
        'spec': specification_version,
        'meson_min_version': meson_min_version,
        'py_major': python_support.major,
        'py_security': '.'.join(map(str, (python_support.major, python_support.security))),
        'py_bugfix2': '.'.join(map(str, (python_support.major, python_support.bugfix2))),
        'py_bugfix1': '.'.join(map(str, (python_support.major, python_support.bugfix1))),
        'py_implementations': implementation_support,
    }
}


def check_for_update(
    current_version: Version, releases: List[str]
) -> None:  # pragma: no cover
    """Issue a warning if installed version of OZI is not up to date."""
    match max(map(parse, releases)):
        case latest if latest > current_version:
            warn(
                'Newer version of OZI available to download on PyPI: '
                'https://pypi.org/project/OZI/',
                RuntimeWarning,
            )
        case latest if latest < current_version:
            print('ok - OZI package is development version.')
        case latest if latest == current_version:
            print('ok - OZI package is up to date.')



# ozi/assets/structure.py
# Part of the OZI Project, under the Apache License v2.0 with LLVM Exceptions.
# See LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""Directory structure expected files"""
required_pkg_info = (
    'Description-Content-Type',
    'Home-page',
    'Metadata-Version',
    'Name',
    'Summary',
    'Version',
    'Author',
    'Author-email',
    'License',
)
root_dirs = ['tests', 'subprojects']
root_files = [
    'README.rst',
    '.gitignore',
    'pyproject.toml',
    'meson.build',
    'meson.options',
    'LICENSE.txt',
    'PKG-INFO',
    'requirements.in',
    'CHANGELOG.md',
]
source_files = [
    'meson.build',
    '__init__.pyi',
    '__init__.py',
    'py.typed',
]
test_files = [
    'meson.build',
]



# ozi/assets/__init__.py
# Part of the OZI Project, under the Apache License v2.0 with LLVM Exceptions.
# See LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""Asset files for python packaging."""
import argparse
import platform
import re
from dataclasses import dataclass
from datetime import date, datetime, timedelta, timezone
from difflib import get_close_matches
from typing import Any, List, Optional, Sequence, Union
from warnings import warn

from pyparsing import Forward, Keyword, Literal, ZeroOrMore, oneOf
from spdx_license_list import LICENSES  # type: ignore

# pyright: reportMissingImports = false
from trove_classifiers import classifiers

required_pkg_info_patterns = (
    'Author',
    'Author-email',
    'Description-Content-Type',
    'Home-page',
    'License',
    'License-Expression',
    'License-File',
    'Metadata-Version',
    'Name',
    'Programming Language :: Python',
    'Summary',
    'Version',
)
metadata_version = '2.1'
implementation_support = ('CPython',)
specification_version = '0.1'
meson_min_version = '1.1.0'
required_root_dist = [
    'README.rst',
    'CHANGELOG.md',
    'pyproject.toml',
    'PKG-INFO',
    'LICENSE.txt',
]
__pymajor, __pyminor, __pypatch = map(int, platform.python_version_tuple())

minor_deprecation = {
    8: date(2024, 10, 1),
    9: date(2025, 10, 1),
    10: date(2026, 10, 1),
    11: date(2027, 10, 1),
    12: date(2028, 10, 1),
    13: date(2029, 10, 1),
}
python3_eol = minor_deprecation.get(__pyminor, date(2008, 12, 3))
current_date = datetime.now(tz=timezone.utc).date()
ozi_support_eol = python3_eol - timedelta(weeks=104)
if current_date > python3_eol:  # pragma: no cover
    raise RuntimeError(
        f'Python {__pymajor}.{__pyminor}.{__pypatch} is not supported as of {python3_eol}.'
    )
elif current_date > ozi_support_eol:  # pragma: no cover
    warn(
        f'Python {__pymajor}.{__pyminor}.{__pypatch} support is deprecated as of {ozi_support_eol}.',
        DeprecationWarning,
    )


@dataclass(frozen=True)
class __PythonSupport:
    """Datatype for OZI Python version support."""

    major: int
    bugfix1: int
    bugfix2: int
    security: int


def __python_support_versions(date: date) -> __PythonSupport:
    """Get the currently supported Python versions from the schedule."""
    return __PythonSupport(
        3,
        *sorted(
            [k for k, v in minor_deprecation.items() if v - timedelta(weeks=104) > date]
        )[:3],
    )


python_support = __python_support_versions(current_date)
python_support_required = {
    ('Classifier', f'Programming Language :: Python :: {str(python_support.major)} :: Only'),
    (
        'Classifier',
        'Programming Language :: Python :: '
        f'{str(python_support.major)}.{str(python_support.security)}',
    ),
    (
        'Classifier',
        'Programming Language :: Python :: '
        f'{str(python_support.major)}.{str(python_support.bugfix2)}',
    ),
    (
        'Classifier',
        'Programming Language :: Python :: '
        f'{str(python_support.major)}.{str(python_support.bugfix1)}',
    ),
}

pep639_spdx = [
    'LicenseRef-Public-Domain',
    'LicenseRef-Proprietary',
]
spdx_exceptions = [
    '389-exception',
    'Asterisk-exception',
    'Autoconf-exception-2.0',
    'Autoconf-exception-3.0',
    'Autoconf-exception-generic',
    'Autoconf-exception-macro',
    'Bison-exception-2.2',
    'Bootloader-exception',
    'Classpath-exception-2.0',
    'CLISP-exception-2.0',
    'cryptsetup-OpenSSL-exception',
    'DigiRule-FOSS-exception',
    'eCos-exception-2.0',
    'Fawkes-Runtime-exception',
    'FLTK-exception',
    'Font-exception-2.0',
    'freertos-exception-2.0',
    'GCC-exception-2.0',
    'GCC-exception-3.1',
    'GNAT-exception',
    'gnu-javamail-exception',
    'GPL-3.0-interface-exception',
    'GPL-3.0-linking-exception',
    'GPL-3.0-linking-source-exception',
    'GPL-CC-1.0',
    'GStreamer-exception-2005',
    'GStreamer-exception-2008',
    'i2p-gpl-java-exception',
    'KiCad-libraries-exception',
    'LGPL-3.0-linking-exception',
    'libpri-OpenH323-exception',
    'Libtool-exception',
    'Linux-syscall-note',
    'LLGPL',
    'LLVM-exception',
    'LZMA-exception',
    'mif-exception',
    'OCaml-LGPL-linking-exception',
    'OCCT-exception-1.0',
    'OpenJDK-assembly-exception-1.0',
    'openvpn-openssl-exception',
    'PS-or-PDF-font-exception-20170817',
    'QPL-1.0-INRIA-2004-exception',
    'Qt-GPL-exception-1.0',
    'Qt-LGPL-exception-1.1',
    'Qwt-exception-1.0',
    'SHL-2.0',
    'SHL-2.1',
    'SWI-exception',
    'Swift-exception',
    'u-boot-exception-2.0',
    'Universal-FOSS-exception-1.0',
    'vsftpd-openssl-exception',
    'WxWindows-exception-3.1',
    'x11vnc-openssl-exception',
]

ambiguous_licenses = [
    'Private',
    'DFSG approved',
    'OSI Approved :: Academic Free License (AFL)',
    'OSI Approved :: Apache Software License',
    'OSI Approved :: Apple Public Source License',
    'OSI Approved :: Artistic License',
    'OSI Approved :: BSD License',
    'OSI Approved :: GNU Affero General Public License v3',
    'OSI Approved :: GNU Free Documentation License (FDL)',
    'OSI Approved :: GNU General Public License (GPL)',
    'OSI Approved :: GNU General Public License v2 (GPLv2)',
    'OSI Approved :: GNU General Public License v3 (GPLv3)',
    'OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)',
    'OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
    'OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
    'OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
    'Public Domain',
]
spdx_disambiguation = [
    ('LicenseRef-Proprietary',),
    (
        'AGPL-3.0-only',
        'AGPL-3.0-or-later',
        'Apache-2.0',
        'Artistic-2.0',
        'BSD-3-Clause',
        'CC-BY-4.0',
        'CC-BY-SA-4.0',
        'EPL-1.0',
        'GPL-2.0-only',
        'GPL-2.0-or-later',
        'GPL-3.0-only',
        'GPL-3.0-or-later',
        'ISC',
        'LGPL-2.1-or-later',
        'LGPL-3.0-only',
        'LGPL-3.0-or-later',
        'MIT',
        'OFL-1.1',
        'WTFPL',
        'Zlib',
    ),
    ('AFL-3.0',),
    ('Apache-2.0',),
    ('APSL-1.0', 'APSL-1.1', 'APSL-1.2', 'APSL-2.0'),
    ('Artistic-2.0',),
    ('0BSD', 'BSD-2-Clause', 'BSD-3-Clause', 'BSD-3-Clause-Clear', 'BSD-4-Clause'),
    ('AGPL-3.0-only', 'AGPL-3.0-or-later'),
    ('GFDL-1.3-only', 'GFDL-1.3-or-later'),
    ('GPL-2.0-only', 'GPL-2.0-or-later', 'GPL-3.0-only', 'GPL-3.0-or-later'),
    ('GPL-2.0-only', 'GPL-2.0-or-later'),
    ('GPL-3.0-only', 'GPL-3.0-or-later'),
    ('LGPL-2.0-only',),
    ('LGPL-2.1-or-later',),
    ('LGPL-3.0-only', 'LGPL-3.0-or-later'),
    ('LGPL-2.1-or-later', 'LGPL-3.0-only', 'LGPL-3.0-or-later'),
    ('LicenseRef-Public-Domain', 'CC0-1.0', 'Unlicense'),
]
spdx_options = dict(zip(ambiguous_licenses, spdx_disambiguation))
spdx_license_expression = Forward()
spdx_license_expression <<= oneOf(
    (pep639_spdx + [lic.id for lic in LICENSES.values() if not lic.deprecated_id])
).set_name('License-ID') + ZeroOrMore(
    (
        Keyword('WITH') + oneOf(spdx_exceptions).set_name('License-Exception-ID')
        | Keyword('AND') + spdx_license_expression
        | Keyword('OR') + spdx_license_expression
    )
) | Literal(
    '('
) + spdx_license_expression + Literal(
    ')'
)
top4 = [
    'MIT License',
    'BSD License',
    'GNU General Public License v3',
    'Apache Software License',
]
framework_prefix = 'Framework :: '
environment_prefix = 'Environment :: '
audience_prefix = 'Intended Audience :: '
status_prefix = 'Development Status :: '
language_prefix = 'Natural Language :: '
license_prefix = 'License :: '
topic_prefix = 'Topic :: '
root_templates = [
    '.gitignore',
    'meson.build',
    'meson.options',
    'PKG-INFO',
    'pyproject.toml',
    'README.rst',
    'LICENSE.txt',
    'requirements.in',
    'CHANGELOG.md',
]
source_templates = [
    'project.name/__init__.py',
    'project.name/__init__.pyi',
    'project.name/meson.build',
    'project.name/py.typed',
]
test_templates = [
    'tests/meson.build',
]
ci_provider_templates = ['github_workflows/ozi.yml']
new_module_templates = ['project.name/new_module.py']
new_test_templates = ['tests/new_test.py']


class CloseMatch(argparse.Action):
    """Special choices action. Warn the user if a close match could not be found."""

    audience = [
        i[len(audience_prefix) :].lstrip()
        for i in classifiers
        if i.startswith(audience_prefix)
    ]
    language = [
        i[len(language_prefix) :].lstrip()
        for i in classifiers
        if i.startswith(language_prefix)
    ]
    framework = [
        i[len(framework_prefix) :].lstrip()
        for i in classifiers
        if i.startswith(framework_prefix)
    ]
    environment = [
        i[len(environment_prefix) :].lstrip()
        for i in classifiers
        if i.startswith(environment_prefix)
    ]
    license = [
        i[len(license_prefix) :].lstrip()
        for i in classifiers
        if i.startswith(license_prefix)
    ]
    license_id = [k for k, v in LICENSES.items() if v.deprecated_id is False]
    license_exception_id = spdx_exceptions
    status = [
        i[len(status_prefix) :].lstrip() for i in classifiers if i.startswith(status_prefix)
    ]
    topic = [
        i[len(topic_prefix) :].lstrip() for i in classifiers if i.startswith(topic_prefix)
    ]

    def __init__(
        self: argparse.Action,
        option_strings: List[str],
        dest: str,
        nargs: Union[int, str, None] = None,
        **kwargs: Any,
    ) -> None:
        """argparse init"""
        if nargs not in [None, '?']:
            raise ValueError('nargs (other than "?") not allowed')

        super().__init__(option_strings, dest, nargs=nargs, **kwargs)  # type: ignore

    def close_matches(
        self: argparse.Action, key: str, value: Sequence[str]
    ) -> Sequence[str]:
        """Get a single close match for a class attribute."""
        if value is None:
            return []  # pragma: defer to good-first-issue
        try:
            value = get_close_matches(value, self.__getattribute__(key), cutoff=0.40)[0]
        except (IndexError, AttributeError):
            warn(
                f'No {key} choice matching "{value}" available.'
                'To list available options:'
                f'$ ozi-new -l {key}',
                RuntimeWarning,
            )
        return value

    def __call__(
        self: argparse.Action,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Union[str, Sequence[str], None],
        option_string: Union[str, None] = None,
    ) -> None:
        """Action business logic."""
        if values is None:
            return
        if option_string is not None:
            key = option_string.lstrip('-').replace('-', '_')
        else:
            key = ''  # pragma: defer to good-first-issue

        if self.nargs == '?':
            matches: List[str] = []
            for v in values:
                v = self.close_matches(key, v)  # type: ignore
                matches += v
            setattr(namespace, self.dest, matches)
        else:
            values = self.close_matches(key, values)  # type: ignore
            setattr(namespace, self.dest, values)


def underscorify(s: str) -> str:
    """Filter to replace non-alphanumerics with underscores."""
    return re.sub('[^0-9a-zA-Z]', '_', s)


def wheel_repr(version: str) -> str:
    """Filter to transform versions of the form "X.Y" into "pyXY"."""
    major, minor = version.split('.')
    return f'py{major}{minor}'


def tap_warning_format(
    msg: str, category: type[Warning], filename: str, lineno: int, line: Optional[str] = None
) -> str:
    """Test Anything Protocol formatted warnings."""
    return f'# {filename}:{lineno}: {category.__name__}\nnot ok - {msg}\n'



# ozi/fix.py
# Part of the OZI Project, under the Apache License v2.0 with LLVM Exceptions.
# See LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""ozi-fix: Project fix script that outputs a meson rewriter JSON array."""
from __future__ import annotations

import argparse
import json
import os
import re
import sys
import warnings
from contextlib import contextmanager, suppress
from dataclasses import asdict, dataclass, field
from email import message_from_file
from email.message import Message
from functools import partial
from pathlib import Path
from runpy import run_module
from typing import (
    Annotated,
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Mapping,
    NoReturn,
    Set,
    Tuple,
    Union,
)
from warnings import warn

from jinja2 import Environment, PackageLoader, Template, select_autoescape
from pyparsing import (
    CaselessKeyword,
    Combine,
    Forward,
    Keyword,
    Literal,
    ParseException,
    ParseResults,
    Suppress,
    White,
    oneOf,
)

from . import metadata
from .assets import (
    python_support_required,
    spdx_license_expression,
    tap_warning_format,
    underscorify,
)
from .assets.structure import (  # noqa: F401
    required_pkg_info,
    root_files,
    source_files,
    test_files,
)

warnings.formatwarning = tap_warning_format  # type: ignore

env = Environment(
    loader=PackageLoader('ozi'),
    autoescape=select_autoescape(),
    enable_async=True,
)
env.filters['underscorify'] = underscorify
env.globals = env.globals | metadata
parser = argparse.ArgumentParser(description=sys.modules[__name__].__doc__, add_help=False)
subparser = parser.add_subparsers(help='source & test fix', dest='fix')
parser.add_argument('target', type=str, help='target OZI project directory')
source_parser = subparser.add_parser(
    'source', aliases=['s'], description='Create a new Python source in an OZI project.'
)
test_parser = subparser.add_parser(
    'test',
    aliases=['t'],
    description='Create a new Python test in an OZI project.',
)
source_parser.add_argument(
    '-a',
    '--add',
    nargs='?',
    action='append',
    default=['ozi.phony'],
    help='add file or dir/ to project',
)
source_parser.add_argument(
    '-r',
    '--remove',
    nargs='?',
    action='append',
    default=['ozi.phony'],
    help='remove file or dir/ from project',
)
source_parser.add_argument(
    '--copyright-head',
    type=str,
    default='',
    help='copyright header string',
    metavar='Part of the NAME project.\\nSee LICENSE...',
)
source_output = source_parser.add_argument_group('output')
source_output.add_argument(
    '--strict',
    default='--no-strict',
    action=argparse.BooleanOptionalAction,
    help='strict mode raises warnings to errors.',
)
source_output.add_argument(
    '-p', '--pretty', action='store_true', help='pretty print JSON output'
)
test_parser.add_argument(
    '-a',
    '--add',
    nargs='?',
    action='append',
    default=['ozi.phony'],
    help='add file or dir/ to project',
)
test_parser.add_argument(
    '-r',
    '--remove',
    nargs='?',
    action='append',
    default=['ozi.phony'],
    help='remove file or dir/ from project',
)
test_parser.add_argument(
    '--copyright-head',
    type=str,
    default='',
    help='copyright header string',
    metavar='Part of the NAME project.\\nSee LICENSE...',
)
test_output = test_parser.add_argument_group('output')
test_output.add_argument(
    '--strict',
    default='--no-strict',
    action=argparse.BooleanOptionalAction,
    help='strict mode raises warnings to errors.',
)
test_output.add_argument(
    '-p', '--pretty', action='store_true', help='pretty print JSON output'
)

helpers = parser.add_mutually_exclusive_group()
helpers.add_argument('-h', '--help', action='help', help='show this help message and exit')
missing_parser = subparser.add_parser(
    'missing',
    aliases=['m'],
    description='Create a new Python test in an OZI project.',
    add_help=False,
)
missing_parser.add_argument(
    '--add',
    nargs='?',
    action='append',
    default=['ozi.phony'],
    help=argparse.SUPPRESS,
)
missing_parser.add_argument(
    '--remove',
    nargs='?',
    action='append',
    default=['ozi.phony'],
    help=argparse.SUPPRESS,
)
missing_parser.add_argument(
    '--strict',
    default='--no-strict',
    action=argparse.BooleanOptionalAction,
    help='strict mode raises warnings to errors, default: --no-strict',
)
missing_parser.add_argument(
    '--run-utility',
    default='--run-utility',
    action=argparse.BooleanOptionalAction,
    help='run-utility mode runs isort, autoflake, and black with OZI standard arguments,'
    ' default: --run-utlity',
)
missing_parser.add_argument(
    '--pretty',
    default='--no-pretty',
    action=argparse.BooleanOptionalAction,
    help='pretty mode outputs indented json, default: --no-pretty',
)


def _str_dict_union(toks: ParseResults) -> Dict[str, str]:
    """Parse-time union of Dict[str, str]."""
    if len(toks) >= 2:
        return toks[0] | toks[1]  # type: ignore
    else:  # pragma: no cover
        return  # type: ignore


sspace = Suppress(White(' ', exact=1))
dcolon = sspace + Suppress(Literal('::')) + sspace
classifier = Suppress(White(' ', min=2)) + Suppress(Literal('Classifier:')) + sspace
pep639_headers = Forward()
license_expression = classifier + (
    Keyword('License-Expression')
    + dcolon
    + Combine(spdx_license_expression, join_string=' ')
).set_parse_action(lambda t: {str(t[0]): str(t[1])})
license_file = classifier + (
    Keyword('License-File') + dcolon + oneOf(['LICENSE', 'LICENSE.txt'])
).set_parse_action(lambda t: {str(t[0]): str(t[1])})
pep639_headers <<= license_expression + license_file

pep639_parse = Suppress(
    Keyword('..') + CaselessKeyword('ozi')
) + pep639_headers.set_parse_action(_str_dict_union).set_name('pep639')


def pkg_info_extra(payload: str, as_message: bool = True) -> Union[Dict[str, str], Message]:
    """Get extra PKG-INFO Classifiers tacked onto the payload by OZI."""
    pep639: Dict[str, str] = pep639_parse.parse_string(payload)[0]  # pyright: ignore
    if as_message:
        msg = Message()
        for k, v in pep639.items():
            msg.add_header('Classifier', f'{k} :: {v}')
        return msg
    else:
        return pep639


def missing_python_support(
    pkg_info: Message, count: int, stdout: Callable[..., None]
) -> Tuple[int, Set[Tuple[str, str]]]:
    """Check PKG-INFO Message for python support."""
    remaining_pkg_info = {(k, v) for k, v in pkg_info.items() if k not in required_pkg_info}
    for k, v in iter(python_support_required):
        if (k, v) in remaining_pkg_info:
            count += 1
            stdout('ok', count, '-', f'{k}:', v)
        else:
            warn(f'{count} - "{v}" MISSING', RuntimeWarning)
    return count, remaining_pkg_info


def missing_ozi_required(
    pkg_info: Message, count: int, stdout: Callable[..., None]
) -> Tuple[int, Any]:
    """Check missing required OZI extra PKG-INFO"""
    count, remaining_pkg_info = missing_python_support(pkg_info, count, stdout)
    remaining_pkg_info.difference_update(set(iter(python_support_required)))
    for k, v in iter(remaining_pkg_info):
        count += 1
        stdout('ok', count, '-', f'{k}:', v)
    try:
        extra_pkg_info = pkg_info_extra(pkg_info.get_payload()).items()
    except ParseException as e:
        extra_pkg_info = {}  # type: ignore
        newline = '\n'
        errstr = str(e).replace(newline, ' ')
        warn(f'{count} - MISSING {errstr}', RuntimeWarning)
    return count, extra_pkg_info


def missing_required(
    target: Path, count: int, stdout: Callable[..., None]
) -> Tuple[int, str, Any]:
    """Find missing required PKG-INFO"""
    with target.joinpath('PKG-INFO').open() as f:
        pkg_info = message_from_file(f)
        count += 1
        stdout('ok', count, '-', 'Parse PKG-INFO')
    for i in required_pkg_info:
        count += 1
        v = pkg_info.get(i, None)
        if v is not None:
            stdout('ok', count, '-', f'{i}:', v)
        else:
            warn(f'{count} - {i} MISSING', RuntimeWarning)
    count, extra_pkg_info = missing_ozi_required(pkg_info, count, stdout)
    name = re.sub(r'[-_.]+', '-', pkg_info.get('Name', str())).lower()
    for k, v in extra_pkg_info:
        count += 1
        stdout('ok', count, '-', f'{k}:', v)
    return count, name, extra_pkg_info


def count_comments(  # noqa: C901
    count: int, lines: List[str], rel_path: Path
) -> int:  # pragma: defer to good-first-issue
    for i, line in enumerate(lines, start=1):
        if s := re.search(r'(.*?noqa:\s[A-Z0-9]*.*?)', line):
            count += 1
            print('#', 'NOQA', f'{str(rel_path)}:{i}', s[0].strip())
        if s := re.search(r'(.*?type:\signore.*?)', line):
            count += 1
            print('#', 'TYPE', f'{str(rel_path)}:{i}', s[0].strip())
        if s := re.search(r'(.*?pragma:\sdefer\sto\s[a-z_-]*.*?)', line):
            count += 1
            print(
                '#',
                'PRAGMA',
                f'{str(rel_path)}:{i}',
                s[0].strip(),
            )
        if s := re.search(r'(.*?pragma:\sno\scover\s.*?)', line):
            count += 1
            print(
                '#',
                'PRAGMA',
                f'{str(rel_path)}:{i}',
                s[0].strip(),
            )
    return count


def missing_required_files(  # noqa: C901
    kind: str,
    target: Path,
    count: int,
    miss_count: int,
    name: str,
    stdout: Callable[..., None],
) -> Tuple[int, int, List[str], List[str]]:
    """Count missing files required by OZI"""
    found_files = []
    miss_count = 0
    mapping = {
        'test': Path('tests'),
        'source': Path(underscorify(name)),
        'root': Path('.'),
    }
    for file in vars(sys.modules[__name__]).get('_'.join([kind, 'files']), []):
        rel_path = mapping.get(kind, Path('.')) / file
        if not target.joinpath(rel_path).exists():
            count += 1
            warn(f'{count} - {rel_path} MISSING', RuntimeWarning)
            miss_count += 1
            continue  # pragma: defer to https://github.com/nedbat/coveragepy/issues/198
        elif str(rel_path).endswith('.py'):
            with open(target.joinpath(rel_path), 'r') as f:
                count_comments(count, f.readlines(), rel_path)
        count += 1
        stdout('ok', count, '-', rel_path)
        found_files.append(file)
    rel_path = mapping.get(kind, Path('.'))
    extra_files = [
        file
        for file in os.listdir(target / rel_path)
        if os.path.isfile(target / rel_path / file)
    ]
    extra_files = list(set(extra_files).symmetric_difference(set(found_files)))
    build_files = []
    for file in extra_files:  # pragma: no cover
        with open(str((rel_path / file).parent / 'meson.build'), 'r') as f:
            for _ in [
                i for i in f.readlines() if re.search(fr'(.*?[\'|"]{file}[\'|"].*?)', i)
            ]:
                count += 1
                build_file = str((rel_path / file).parent / 'meson.build')
                stdout(
                    'ok',
                    count,
                    '-',
                    f'{build_file}:',
                    str(rel_path / file),
                )
        build_files += [str(rel_path / file)]
        if str(file).endswith('.py') and str(rel_path / file) not in build_files:
            build_file = str(rel_path / 'meson.build')
            warn(
                f'{count} - MISSING {build_file}: {str(rel_path / file)}',
                RuntimeWarning,
            )
            print(f'# SKIP {str(rel_path / file)}')
        elif str(file).endswith('.py'):
            with open(target.joinpath(rel_path) / file, 'r') as g:
                count_comments(count, g.readlines(), rel_path / file)
    return count, miss_count, found_files, extra_files


def report_missing(
    target: Path, stdout: Callable[..., None] = print
) -> Union[
    Tuple[str, Message, List[str], List[str], List[str]], Tuple[None, None, None, None, None]
]:
    """Report missing OZI project files
    :param target: Relative path to target directory.
    :return: Normalized Name, PKG-INFO, found_root, found_sources, found_tests
    """
    target = Path(target)
    miss_count = 0
    count = 0
    name = None
    pkg_info = None
    extra_pkg_info: Dict[str, str] = {}
    try:
        count, name, extra_pkg_info = missing_required(target, count, stdout)
    except FileNotFoundError:
        name = ''
        warn(f'{count} - PKG-INFO MISSING', RuntimeWarning)
    count, miss_count, found_source_files, extra_source_files = missing_required_files(
        'source', target, count, 0, name, stdout
    )
    count, miss_count, found_test_files, extra_test_files = missing_required_files(
        'test', target, count, miss_count, name, stdout
    )
    count, miss_count, found_root_files, extra_root_files = missing_required_files(
        'root', target, count, miss_count, name, stdout
    )
    all_files = (
        ['PKG-INFO'],
        extra_pkg_info,
        found_root_files,
        found_source_files,
        found_test_files,
        extra_root_files,
        extra_source_files,
        extra_test_files,
    )
    try:
        sum(map(len, all_files))
    except TypeError:  # pragma: no cover
        warn('Bail out! MISSING required files or metadata.')
        return (None, None, None, None, None)
    stdout(f'1..{count+miss_count}')
    return name, pkg_info, found_root_files, found_source_files, found_test_files  # type: ignore


@dataclass
class RewriteCommand:  # pragma: defer to meson
    """Meson rewriter command input"""

    active: bool = field(repr=False, default_factory=bool, init=False)
    type: str = 'target'
    target: str = ''
    operation: str = ''
    sources: List[str] = field(default_factory=list)
    subdir: str = ''
    target_type: str = ''

    def add(
        self: RewriteCommand, mode: str, kind: str, source: str
    ) -> Dict[str, Union[str, List[str]]]:
        """Add sources and tests to an OZI project."""
        self.sources += [source]
        self.operation = 'src_add'
        return self._body(mode, kind)

    def rem(
        self: RewriteCommand, mode: str, kind: str, source: str
    ) -> Dict[str, Union[str, List[str]]]:
        """Add sources and tests to an OZI project."""
        self.sources += [source]
        self.operation = 'src_rem'
        return self._body(mode, kind)

    def _body(
        self: RewriteCommand, mode: str, kind: str
    ) -> Dict[str, Union[str, List[str]]]:
        """The body of the add/rem functions"""
        self.active = True
        target = mode + '_' + kind
        mode_set = self.target == target or self.target == ''
        if mode_set:
            self.target = target
        else:
            raise ValueError(f'target already set to {self.target}')
        return asdict(self)


@dataclass
class Rewriter:
    """Container for Meson rewriter commands."""

    target: str
    name: str
    fix: str
    commands: List[Dict[str, str]] = field(default_factory=list)
    path_map: Mapping[str, Callable[[str], Path]] = field(init=False)
    base_templates: Dict[
        Annotated[str, 'fix'], Annotated[Template, 'base_template']
    ] = field(init=False)

    def __post_init__(self: Rewriter) -> None:
        """Setup the path_map"""
        self.path_map = {
            'source': partial(Path, self.target, self.name),
            'test': partial(Path, self.target, 'tests'),
            'root': partial(Path, self.target),
        }
        self.base_templates = {
            'root': env.get_template('tests/new_test.py.j2'),
            'source': env.get_template('project.name/new_module.py.j2'),
            'test': env.get_template('tests/new_test.py.j2'),
        }

    def find_user_templates(self: Rewriter, file: str) -> str | None:
        try:
            with open(Path(self.target, 'templates', self.fix, file)) as template:
                user_template = template.read()
        except OSError:
            user_template = None
        return user_template

    def _add(  # noqa: C901
        self: Rewriter,
        child: Path,
        file: str,
        cmd_files: RewriteCommand,
        cmd_children: RewriteCommand,
    ) -> Tuple[RewriteCommand, RewriteCommand]:
        """Add items to OZI Rewriter"""
        if self.fix not in ['source', 'test', 'root']:
            warn('Invalid fix mode nothing will be added.', RuntimeWarning)
        else:
            if file.endswith('/'):
                child.mkdir(parents=True)
                parent = file.rstrip('/')
                heirs = parent.split('/')
                if len(heirs) > 1:
                    warn('Nested folder creation not supported.', RuntimeWarning)
                else:
                    with open((child / 'meson.build'), 'x') as f:
                        f.write(env.get_template('new_child.j2').render(parent=parent))
                if self.fix == 'source':
                    if len(heirs) > 1:
                        warn('Nested folder creation not supported.', RuntimeWarning)
                    else:
                        with open(
                            (
                                self.path_map.get(self.fix, partial(Path))(*heirs)
                                / '__init__.py'
                            ),
                            'x',
                        ) as f:
                            f.write(
                                env.get_template('project.name/__init__.py.j2').render(
                                    user_template=self.find_user_templates(file)
                                )
                            )
                        cmd_children.add(self.fix, 'children', parent)
                else:
                    cmd_children.add(self.fix, 'children', parent)
            elif file.endswith('.py'):
                child.write_text(
                    self.base_templates.get(
                        self.fix,
                        self.base_templates.setdefault(
                            self.fix, env.get_template('project.name/new_module.py.j2')
                        ),
                    ).render(user_template=self.find_user_templates(file))
                )
                cmd_files.add(self.fix, 'files', str(Path(file)))
            else:
                child.touch()
                cmd_files.add(self.fix, 'files', str(Path(file)))
        return cmd_files, cmd_children

    def __iadd__(self: Rewriter, other: List[str]) -> Rewriter:
        """Add a list of paths"""
        cmd_files = RewriteCommand()
        cmd_children = RewriteCommand()

        for file in other:
            child = self.path_map.get(self.fix, partial(Path))(file)
            cmd_files, cmd_children = self._add(child, file, cmd_files, cmd_children)
        if cmd_files.active:
            self.commands += [{k: v for k, v in asdict(cmd_files).items() if k != 'active'}]
        if cmd_children.active:
            self.commands += [
                {k: v for k, v in asdict(cmd_children).items() if k != 'active'}
            ]
        return self

    def _sub(
        self: Rewriter,
        child: Path,
        file: str,
        cmd_files: RewriteCommand,
        cmd_children: RewriteCommand,
    ) -> Tuple[RewriteCommand, RewriteCommand]:
        """Remove items from OZI Rewriter"""
        if file.endswith('/'):
            cmd_children.rem(self.fix, 'children', str(child / 'meson.build'))
        elif file.endswith('.py'):
            cmd_files.rem(self.fix, 'files', str(Path(file)))
        else:
            cmd_files.rem(self.fix, 'files', str(Path(file)))
        return cmd_files, cmd_children

    def __isub__(self: Rewriter, other: List[str]) -> Rewriter:
        """Remove a list of paths"""
        cmd_files = RewriteCommand()
        cmd_children = RewriteCommand()
        for file in other:
            child = self.path_map.get(self.fix, partial(Path))(file)
            cmd_files, cmd_children = self._sub(child, file, cmd_files, cmd_children)
            self.__rm_dir(file, child)
        if cmd_files.active:
            self.commands += [{k: v for k, v in asdict(cmd_files).items() if k != 'active'}]
        if cmd_children.active:
            self.commands += [
                {k: v for k, v in asdict(cmd_children).items() if k != 'active'}
            ]
        return self

    def __rm_dir(self: Rewriter, file: str, child: Path) -> None:
        """Try to remove a directory if empty."""
        if file.endswith('/'):
            try:
                child.rmdir()
            except OSError:
                warn(
                    f'Could not remove non-empty or non-existing {self.fix}: "{child}".',
                    RuntimeWarning,
                )


def preprocess(project: argparse.Namespace) -> argparse.Namespace:
    """Remove phony arguments, check target exists and is a directory, set missing flag."""
    project.missing = project.fix == 'missing'
    project.target = Path(project.target).absolute()
    if not project.target.exists():
        warn(f'Bail out! target: {project.target}\ntarget does not exist.', RuntimeWarning)
    elif not project.target.is_dir():
        warn(
            f'Bail out! target: {project.target}\ntarget is not a directory.', RuntimeWarning
        )
    project.add.remove('ozi.phony')
    project.add = list(set(project.add))
    project.remove.remove('ozi.phony')
    project.remove = list(set(project.remove))
    return project


@contextmanager
def redirect_argv(*args: str) -> Generator[None, None, None]:  # pragma: no cover
    argv = sys.argv[:]
    sys.argv = list(args)
    yield
    sys.argv = argv


def run_utility(name: str, *args: str) -> None:  # pragma: no cover
    with redirect_argv(name, *args):
        print('# run-utility:', *sys.argv)
        with suppress(SystemExit):
            run_module(name)


def main() -> NoReturn:  # pragma: no cover
    """Main ozi.fix entrypoint."""
    project = preprocess(parser.parse_args())
    env.globals = env.globals | {
        'project': vars(project),
        **metadata,
    }
    name, *_ = report_missing(
        project.target, stdout=print if project.missing else lambda *_: None  # type: ignore
    )

    if name is None:
        exit(1)

    if hasattr(project, 'run_utility') and project.run_utility:
        run_utility('isort', '-q', str(project.target))
        for i in [underscorify(name), 'tests']:
            run_utility('autoflake', '-i', str(project.target / i / '**' / '*.py'))
            run_utility('black', '-q', '-S', str(project.target / i))

    project.name = underscorify(name)
    project.license_file = 'LICENSE.txt'
    project.copyright_head = '\n'.join(
        [
            f'Part of {name}.',
            f'See {project.license_file} in the project root for details.',
        ]
    )
    rewriter = Rewriter(project.target, project.name, project.fix)
    rewriter += project.add
    rewriter -= project.remove
    if not project.missing:
        print(json.dumps(rewriter.commands, indent=4 if project.pretty else None))
    exit(0)


if __name__ == '__main__':
    main()
